--[[Usage
	=====

	Create a menu
	=============
	menu=DMMenuCreate( attachTo )

		Creates an empty menu and returns it as a variable.
		An undefined number of menus can be created.

		attachTo	Optional. On-screen element to attach the menu to.


	Create a dropdown-box menu
	==========================
	menu=DMDropDownCreate( attachTo, left )

		Creates an empty dropdown menu and returns it as a variable.
		The attached element will be given a button that will open the
		menu when clicked, to emulate the operational mode of a dropdown
		box. 
		An undefined number of menus can be created.

		attachTo	Mandatory. On-screen element to attach the menu to.
		left		Optional. When true, the button will be positioned at
					the left edge of the attached element as opposed to
					the normal right edge.


	Add simple entries to a menu
	============================
	menu:Add( "Text", function )

		Adds a clickable menu entry that will call the function when clicked.

		"Text"		Mandatory. Menu text for the entry.
		function	Optional. Function to call when entry is clicked.


	Add complex entries to a menu
	=============================
	menu:Add( "Text", function, mode, tooltip, checked, inactive, indent, icon, group )

		Adds a clickable menu entry that will call the function when clicked.

		"Text"		Mandatory. Menu text for the entry.
		function	Optional. Function to call when entry is clicked.
		mode		Optional. Entry mode: DDT_XXXX
		tooltip		Optional. Tooltip string or table of strings for hover
					tooltip display.
		checked		Optional. When set, shows the entry as "checked".
		inactive	Optional. When set, the entry can not be clicked.
		indent		Optional. A visual indent of the text in small tabulators.
		icon		Optional. A filename for an image to use as entry icon.
		group		Optional. Grouping for radio buttons.

		DDT_XXXX:	DDT_TEXT	- plain text only - unclickable
					DDT_BUTTON	- simple clickable button
					DDT_CHECK	- checkbox
					DDT_RADIO	- radio button
					DDT_MENU	- sub-menu


	Display menu
	============
	menu:Show();

		Shows the menu on-screen.


	Clear menu contents
	===================
	menu:Clear();

		Clear the menu.


	Change status textures
	======================
	menu:Texture( usage, texture, alpha )

		Change the menu's textures.

		usage		Mandatory. String that specifies texture usage. Any
					names allowed by the game engine can be used, like
					"Normal", "Pushed", "Checked", "Highlight", "Disabled".
		texture		Optional. The texture file to use.
		alpha		Optional. The transparency for the texture.


	Change texture colorization
	===========================
	menu:Color( usage, r, g, b, a )
	menu:Color( usage, color )

		Change the menu's texture colour weight. The applied colour is a
		change of the texture colours, so a red colour of 1 will display
		the red of the texture as it is defined by the texture. A red colour
		of 0.5 will display all red components at half intensity, and so on.

		usage		Mandatory. String that specifies texture usage. Any
					names allowed by the game engine can be used, like
					"Normal", "Pushed", "Checked", "Highlight", "Disabled".
		r g b a		Mandatory. Colour coding to apply to the texture.
		color		Mandatory. Colour code string to apply to the texture
					in the game format "#AARRGGBB" or "AARRGGBB".

]]


if ((DuckMod_Menu_Version or 0)<1.08) then			-- This version
	DuckMod_Menu_Version=1.08;

-- types
  DDT_TEXT=0;	-- plain text only
DDT_BUTTON=1;	-- simple clickable button
 DDT_CHECK=2;	-- checkbox
 DDT_RADIO=3;	-- radio button
--  DDT_EDIT=4;	-- edit control
  DDT_MENU=5;	-- sub-menu
-- states
     DDS_NONE=0;
 DDS_INACTIVE=1;
-- colours
DDC_NORMAL="|cFFFFFFFF";
DDC_HEADER="|cFF8080FF";

if (not DMMenu) then DMMenu={}; end					-- If it exists, just overlay
if (not DMMenu.Item) then DMMenu.Item={}; end		-- If it exists, just overlay
if (not DMMenu.Entry) then DMMenu.Entry={}; end		-- If it exists, just overlay

local function copytable(t,new)
	if (not t and not new) then return nil; end
	if (not t) then t={}; end
	if (not new) then new={}; end
	for i,v in pairs(t) do if (type(v)=="table") then new[i]=copytable(v,new[i]); else new[i]=v; end end
	return new;
end

local function SetMenuFunctions(menu)
	menu=menu or {};
	menu.Add=DMMenu.Add;
	menu.Header=DMMenu.Header;
	menu.Separator=DMMenu.Separator;
	menu.Text=DMMenu.Text;
	menu.Button=DMMenu.Button;
	menu.Checkbox=DMMenu.Checkbox;
	menu.Radiobutton=DMMenu.Radiobutton;
	menu.Clear=DMMenu.Clear;
	menu.Show=DMMenu.Show;
	menu.Color=DMMenu.Colour;
	menu.Texture=DMMenu.Texture;
	menu.Submenu=DMMenu.Submenu;
	menu.Background=DMMenu.Background;
	menu.Edge=DMMenu.Edge;
	return menu;
end

local function SetEntryFunctions(entry)
	if (not entry) then return; end
	entry.AddEdit=DMMenu.Entry.AddEdit;
end

function DMDropDownCreate(hang,leftedge)
	if (not hang) then return nil; end
	local menu=DMMenuCreate(hang,("LEFT" and leftedge) or not leftedge);	-- shove "LEFT" or true in there
	if (menu.DropDown) then
		local createDD=menu.Hang:GetName().."_AddedDropDownMenuActivatorButton";
		createDD=_G[createDD] or CreateFrame("Button",createDD,menu.Hang,"UIPanelButtonTemplate");
		createDD:ClearAllPoints();
		if (menu.DropDown=="LEFT") then createDD:SetPoint("RIGHT",hang,"LEFT");		-- on the left edge
		else createDD:SetPoint("LEFT",hang,"RIGHT"); end							-- on the right edge
		createDD:SetHeight(hang:GetHeight());
		createDD:SetWidth(10);
		_G[createDD:GetName().."Text"]:SetText("!");
		createDD:SetScript("PreClick",function () DMMenu.Show(menu); end);		-- in-combat safe
		createDD:Show();
	end
	return menu;
end

function DMMenuCreate(hang,createDD)
	return SetMenuFunctions({
		DropDown=createDD,
		Hang=hang,
		Timer=0.5,
		Group=-1111111,
		Settings={
			Colour={ Pushed={r=0,g=1,b=0}, Checked={r=0,g=0,b=1}, Highlight={r=0,g=1,b=0}, },
			Texture={
				Normal="",
				Pushed="Interface\\Buttons\\UI-Listbox-Highlight2",
				Checked="Interface\\Buttons\\UI-Listbox-Highlight2",
				Highlight="Interface\\Buttons\\UI-Listbox-Highlight",
				Disabled="",
			},
			alphaMode={ Checked="ADD", Highlight="ADD", },
		},
		bd={
			bgFile="TILESET\\GENERIC\\Black",
			edgeFile="TILESET\\GENERIC\\Charcoal",
			edgeSize=1,
			insets={left=1,right=1,top=1,bottom=1,},
		},
	});
end

function DMMenu.Header(menu,text) menu:Add(text,nil,DDT_TEXT); end
function DMMenu.Separator(menu) menu:Add(" ",nil,DDT_TEXT); end
function DMMenu.Text(menu,text) return menu:Add(text,nil,DDT_TEXT); end																									-- DDT_TEXT=0;	-- plain text only
function DMMenu.Button(menu,text,func,tooltip,inactive,indent,icon,group) return menu:Add(text,func,DDT_BUTTON,nil,tooltip,inactive,indent,icon,group); end					-- DDT_BUTTON=1;	-- simple clickable button
function DMMenu.Checkbox(menu,text,func,checked,tooltip,inactive,indent,icon,group) return menu:Add(text,func,DDT_CHECK,checked,tooltip,inactive,indent,icon,group); end		-- DDT_CHECK=2;	-- checkbox
function DMMenu.Radiobutton(menu,text,func,checked,tooltip,inactive,indent,icon,group) return menu:Add(text,func,DDT_RADIO,checked,tooltip,inactive,indent,icon,group); end	-- DDT_RADIO=3;	-- radio button
function DMMenu.Submenu(menu,text) return menu:Add(text,nil,DDT_MENU); end				-- DDT_MENU=5;	-- sub-menu


-- "AddEdit" supports the same parameters as "Add", with the difference
-- that "text" is replaced by:
--   label   : label for the entry, string - left of the edit box
--   init    : initial contents of the edit box
--   width   : width of the edit box in screen units
--   unit    : data unit, string - right of the edit box

-- Add an entry to the list
--   self    : The frame
--   text    : Text of the entry to add or complete entry table
--   func    : click-function
--   mode    : Item type
--   tooltip : hover
--   checked : bool
--   inactive: bool
--   indent  : Indentation steps
--   icon    : item icon
--   group   : item group for radio
function DMMenu.Add(menu,text,func,mode,checked,tooltip,inactive,indent,icon,group)
	local buf;
	if (type(text)=="table") then buf=copytable(text);
	else buf={text=text,func=func,mode=mode,tooltip=tooltip,checked=checked,inactive=inactive,indent=indent,icon=icon,group=group or menu.Group}; end
	SetEntryFunctions(buf);
	if (buf.mode==DDT_MENU) then
		buf.checked=nil;
		buf.Menu=DMMenuCreate();
--		SetMenuFunctions();	-- create sub-menu with method for next level
		if (not icon) then buf.icon="Interface\\ChatFrame\\ChatFrameExpandArrow"; end
	else
		if (buf.mode==DDT_TEXT) then buf.text=DDC_HEADER..buf.text; else buf.text=DDC_NORMAL..buf.text; end
		buf.Menu=nil;					-- in case reused table is provided
	end
	table.insert(menu,buf);
	return buf.Menu or buf;				-- self if not menu, handle if menu
end

function DMMenu.Clear(menu)
	for i in ipairs(menu) do menu[i]=nil; end end

function DMMenu.Show(menu)
	local frame=DMMenu:Create(menu,nil);
	frame.Settings=copytable(menu.Settings);
	frame.ddTimer=menu.Timer;
	for _,item in ipairs(menu) do frame:AddItem(item); end
end

function DMMenu.SetFuncs(frame)
	frame.AddItem=DMMenu.frame_AddItem;
	frame.Redraw=DMMenu.frame_Redraw;
	frame.ReAlign=DMMenu.frame_ReAlign;
end

function DMMenu.Entry.AddEdit(entry,text,size,unit,lower,upper)
	entry.Edit=entry.Edit or {}; wipe(entry.Edit);
	entry.Edit.text=text;
	entry.Edit.size=size;
	entry.Edit.unit=unit;
	entry.Edit.lower=lower;
	entry.Edit.upper=upper;
end

function DMMenu.Item.SetButtonText(button,text) _G[button:GetName().."Text"]:SetText(text); end
function DMMenu.Item.GetButtonText(button) return _G[button:GetName().."Text"]:GetText(); end
function DMMenu.Item.SetButtonTextColor(button,r,g,b,a) _G[button:GetName().."Text"]:SetTextColor(r,g,b,a); end
function DMMenu.Item.SetIcon(button,image) _G[button:GetName().."IconTexture"]:SetTexture(image); button.IconSize=((image and 17) or 0); end
function DMMenu.Item.EmptyClick() end;

function DMMenu:Create(menu,level,levelAnchor)
	local hang=menu.Hang;
	level=level or 1;
	if (not DMMenu.Frames) then DMMenu.Frames={}; end
	if (not DMMenu.Frames[level]) then
		DMMenu.Frames[level]=DMMenu.CreateFrame("Frame","DMMenuCreatedFrame"..level,UIParent,"DMMENU");
		tinsert(UISpecialFrames,"DMMenuCreatedFrame"..level);
		DMMenu.Frames[level]:SetScript("OnUpdate",DMMenu.OnUpdate);
		DMMenu.Frames[level]:SetScript("OnEnter",function() end);
	end
	DMMenu.Frames[level]:ClearAllPoints();
	local parent=(WorldMapDetailFrame:IsVisible() and WorldMapDetailFrame) or UIParent;
	if (hang) then
		DMMenu.Frames[level]:SetParent(hang);
		DMMenu.Frames[level]:SetPoint("TOPLEFT",hang,"BOTTOMLEFT");
		DMMenu.Frames[level].ddPoint={ "TOPLEFT",hang,"BOTTOMLEFT",0,0 };
		DMMenu.Frames[level].MinWidth=hang:GetWidth();
	elseif (level>1) then
		DMMenu.Frames[level]:SetParent(DMMenu.Frames[level-1]);
		DMMenu.Frames[level]:SetPoint("TOPLEFT",levelAnchor or DMMenu.Frames[level-1],"TOPRIGHT");
		DMMenu.Frames[level].ddPoint={ "TOPLEFT",levelAnchor or DMMenu.Frames[level-1],"TOPRIGHT",0,0 };
		DMMenu.Frames[level].MinWidth=10;
	else
		local x,y=GetCursorPosition();
		local scale=parent:GetEffectiveScale();
		x=x/scale; y=y/scale;
		DMMenu.Frames[level]:SetParent(parent);
		DMMenu.Frames[level]:SetPoint("TOPLEFT",parent,"BOTTOMLEFT",x,y);
		DMMenu.Frames[level].ddPoint={ "TOPLEFT",parent,"BOTTOMLEFT",x,y };
		DMMenu.Frames[level].MinWidth=10;
	end
	DMMenu.Frames[level]:SetBackdrop(menu.bd);
	_G[DMMenu.Frames[level]:GetName().."_Entry1"]:ClearAllPoints();
	_G[DMMenu.Frames[level]:GetName().."_Entry1"]:SetPoint("TOPLEFT",DMMenu.Frames[level],"TOPLEFT",menu.bd.insets.left,-menu.bd.insets.top);
	_G[DMMenu.Frames[level]:GetName().."_Entry1"]:SetPoint("RIGHT",DMMenu.Frames[level],"RIGHT",-menu.bd.insets.right,0);
	DMMenu.Frames[level].elapsed=3;				-- just so it doesn't crap out on some nil-math later
	DMMenu.Frames[level].ddTimer=0.5;			-- just so it doesn't crap out on some nil-math later
	DMMenu.Frames[level].InOut={};
	DMMenu.Frames[level].Menu={};
	DMMenu.Frames[level].ddEntries=0;
	DMMenu.Frames[level].ddLevel=level;
	DMMenu.Frames[level].ddBounds={ L=parent:GetLeft(), R=parent:GetRight(), T=parent:GetTop(), B=parent:GetBottom(), };
	local i=level; repeat DMMenu.Frames[i].ddAligned=nil; i=i+1; until (not DMMenu.Frames[i]);	-- set this and all following as "not aligned yet"
	i=level+1; while(DMMenu.Frames[i]) do DMMenu.Frames[i].elapsed=-1; DMMenu.Frames[i]:Hide(); i=i+1; end	-- hide all following
	return DMMenu.Frames[level];
end

function DMMenu.frame_AddItem(dd,item)
	local position=nil;
	dd.ddEntries=dd.ddEntries+1;			-- one more in list
	item.text=item.text or "";
	item.mode=item.mode or DDT_BUTTON;
	item.state=item.state or DDS_NONE;
	item.indent=item.indent or 0;
	item.func=item.func or DMMenu.Item.EmptyClick;
	item.group=item.group or -1111111;

	-- create item
	local ddName=dd:GetName();
	local button=_G[ddName.."_Entry"..dd.ddEntries];
	if (not button) then
		button=DMMenu.CreateFrame("CheckButton",ddName.."_Entry"..dd.ddEntries,dd,"DMMENUItem");
		button:ClearAllPoints();
		button:SetPoint("TOPLEFT",_G[ddName.."_Entry"..(dd.ddEntries-1)],"BOTTOMLEFT",0,-dd.ddSpacing);
		button:SetPoint("RIGHT",_G[ddName.."_Entry"..(dd.ddEntries-1)],"RIGHT");
	end
	button.Extra=button.Extra or {};
	if (item.inactive) then button:Disable(); else button:Enable(); end
	if (item.Edit) then					-- menu entry has an edit section
		if (not button.EditBox) then
			button.EditBoxUnit=button:CreateFontString();
			button.EditBoxUnit:SetFontObject("GameFontHighlight");
			button.EditBox=DMMenu.CreateFrame("EditBox",nil,button);
			button.EditBox:SetFontObject("GameFontHighlight");
			button.EditBox:SetJustifyH("RIGHT");
			button.EditBox:SetJustifyV("CENTER");
			button.EditBox:SetTextColor(.2,1,.7);
			button.EditBox:SetAutoFocus(false);
			button.EditBox:SetTextInsets(0,0,0,0);
			button.EditBox:SetScript("OnEnter",function() DMMenu.ItemHover(button,true); SetCursor("INTERACT_CURSOR"); end);
			button.EditBox:SetScript("OnLeave",function() DMMenu.ItemHover(button,nil); SetCursor("POINT_CURSOR"); end);
			button.EditBox:SetScript("OnMouseWheel",function(edit,delta)
						local val=tonumber(edit:GetText()); if (not val) then return; end
						if ((not item.Edit.lower and not item.Edit.upper) or ((item.Edit.lower and val+delta>=item.Edit.lower) and (item.Edit.upper and val+delta<=item.Edit.upper))) then
							val=val+delta;
							edit:SetText(tostring(val));
							button:Clicked(button.ItemChecked,tostring(val));
						end
					end);
			button.EditBox:SetScript("OnEnterPressed",function(edit)
						edit:ClearFocus(); button:GetParent().elapsed=.1;
						local val=button.EditBox:GetText();
						if (type(item.Edit.lower)=="number") then val=tonumber(val) or 0; end
						if ((not item.Edit.lower and not item.Edit.upper) or ((item.Edit.lower and val>=item.Edit.lower) and (item.Edit.upper and val<=item.Edit.upper))) then
							button:Clicked(button.ItemChecked,val);
						end
					end);						-- self - save entry, enable extremely short menu timeout
			button.EditBox:SetScript("OnEscapePressed",function(edit) edit:SetText(edit._default_text); edit:ClearFocus(); end);	-- self - restore original entry, unfocus
			button.EditBox:SetScript("OnTabPressed",function(edit) edit:ClearFocus(); end);		-- self - unfocus
			button.EditBox:SetScript("OnEditFocusGained",function(edit) button:GetParent().NoTimeout=true; edit._default_text=edit:GetText(); edit:HighlightText(0); end);	-- self - disable menu timeout
			button.EditBox:SetScript("OnEditFocusLost",function() button:GetParent().NoTimeout=nil; end);		-- self - enable menu timeout
		end
		if (type(item.Edit.size)~="number") then item.Edit.size=30; end
		button.EditBoxUnit:ClearAllPoints();
		button.EditBoxUnit:SetPoint("TOPRIGHT",button,"TOPRIGHT",-((item.icon and 24) or 0),0);
		button.EditBoxUnit:SetPoint("BOTTOM",button,"BOTTOM");
		button.EditBoxUnit:SetText(item.Edit.unit or "");
		button.EditBoxUnit:Show();
		button.EditBox:ClearAllPoints();
		button.EditBox:SetPoint("RIGHT",button.EditBoxUnit,"LEFT",-2,0);
		button.EditBox:SetPoint("BOTTOM",button,"BOTTOM");
		button.EditBox:SetPoint("TOP",button,"TOP");
		button.EditBox:SetWidth(item.Edit.size);
		button.EditBox:SetText(item.Edit.text or "");
		button.EditBox:Show();
		button.Extra.Edit=true;
	else
		if (button.EditBox) then button.EditBoxUnit:Hide(); button.EditBox:Hide(); end
		button.Extra.Edit=nil;
	end
	button:SetText(item.text);
	button:SetIcon(item.icon);
	button.Clicked=item.func;
	button.ItemMode=item.mode;
	button.ItemState=item.state;
	button.ItemChecked=item.checked;
	button.ItemInactive=item.inactive;
	button.TextIndent=item.indent;
	button.Tooltip=item.tooltip;
	button.Level=dd.ddLevel;
	button.ItemGroup=item.group;
	button.Menu=item.Menu;
	if (dd.GetParent) then button.Master=dd:GetParent(); else button.Master=nil; end	-- widget owning dropdown

	dd:Redraw();
	dd:Show();
	dd.elapsed=3;
	dd:SetFrameStrata("DIALOG");

	return button;
end

function DMMenu.frame_Redraw(dd)
	if (not dd.ddEntries or dd.ddEntries<1) then return; end
	local set=DMMenu.Frames[1].Settings;
	local item=dd:GetName().."_Entry";	-- base item name
	local mw=dd.MinWidth;

	local i=1;
	while(i<=dd.ddEntries) do
		local button=_G[item..i]; if (not button) then break; end
		local width=3+button.TextIndent*button.IndentSize+_G[item..i.."Text"]:GetStringWidth()+button.IconSize+3;	-- left edge (from indent)+text width+right edge (from xml)
		if (button.Extra.Edit) then
			width=width+button.EditBox:GetWidth()+button.EditBoxUnit:GetWidth()+4;	-- all standard + edit + unit + spacers
		end
		if (width>mw) then mw=width; end
		i=i+1;
	end
	dd:SetWidth(mw+2);		-- minimum width + edge-spacer

	i=1;
	while(i<=dd.ddEntries) do
		local button=_G[item..i];
		if (not button) then break; end
		local posL=3+button.TextIndent*button.IndentSize;
		_G[item..i.."Text"]:SetPoint("LEFT","$parent","LEFT",posL,0);

		if (button.ItemState==DDS_INACTIVE) then button:Disable();
		else button:Enable(); end

		if (button.ItemMode==DDT_TEXT) then
			button:SetHighlightTexture("");
			button:SetPushedTexture("");
			button:SetChecked(nil);
		else
			for c,cd in pairs(set.Texture) do button["Set"..c.."Texture"](button,cd,set.alphaMode[c],set.alphaMode[c]); end
			for c,cd in pairs(set.Colour) do button["Get"..c.."Texture"](button):SetVertexColor(cd.r,cd.g,cd.b,cd.a); end
			button:SetChecked(button.ItemChecked);
		end
		button:Show();
		if (button.EditBox) then button.EditBox:ClearFocus(); end
		i=i+1;
	end
	dd:SetHeight(dd.ddEntries*17+(dd.ddEntries-1)*1+2);			-- button height, spacers, edge-spacer
	-- hide surplus
	while(_G[item..i]) do _G[item..i]:Hide(); i=i+1; end
end

function DMMenu.Background(menu,image,tiled,l,r,t,b)
	menu.bd=menu.bd or {};
	menu.bd.bgFile=image;
	menu.bd.tile=(type(tiled)=="number" and tiled>0) or nil;
	menu.bd.tileSize=tiled;
	menu.bd.insets={left=l or 0,right=r or 0,top=t or 0, bottom=b or 0};
end

function DMMenu.Edge(menu,image,size)
	menu.bd=menu.bd or {};
	menu.bd.edgeFile=image;
	menu.bd.edgeSize=size;
end

function DMMenu.Colour(menu,what,r,g,b,a)
	if (type(r)=="string") then
		if (r:sub(1,1)=="#") then r=r:sub(2); end		-- remove number sign
		if (r:len()~=8) then return; end
		a=tonumber(r:sub(1,2),16)/255; b=tonumber(r:sub(7,8),16)/255; g=tonumber(r:sub(5,6),16)/255; r=tonumber(r:sub(3,4),16)/255;
	end
	if (not r and not g and not b) then r,g,b=1,1,1; end
	what=what:sub(1,1):Upper()..what:sub(2):Lower();	-- capitalise
	menu.Settings.Colour[what]={r=r or 0,g=g or 0,b=b or 0,a=a};		-- no forced 'a'
end

function DMMenu.Texture(menu,what,tex,alpha)
	tex=tex or "";
	what=what:sub(1,1):Upper()..what:sub(2):Lower();	-- capitalise
	menu.Settings.Texture[what]=tex;
	menu.Settings.alphaMode[what]=alpha;
end

-- called by XML
function DMMenu.ItemClicked(button)
	if (button.ItemMode==DDT_MENU) then button:SetChecked(nil); return;			-- has no click interface
	elseif (button.ItemMode==DDT_TEXT) then button:SetChecked(nil); return;		-- has no click interface
	elseif (button.ItemMode==DDT_BUTTON) then button.ItemChecked=nil;			-- has no check functionality
	elseif (button.ItemMode==DDT_CHECK) then button.ItemChecked=(not button.ItemChecked) or nil;	-- toggle
	elseif (button.ItemMode==DDT_RADIO) then
		local i,item=1,button:GetParent():GetName().."_Entry";					-- grab parent's base child naming
		while(_G[item..i]) do if (_G[item..i].ItemGroup==button.ItemGroup) then _G[item..i].ItemChecked=nil; end i=i+1; end	-- uncheck all same-grouped
		button.ItemChecked=true;												-- check myself
	else return; end															-- some unknown mode provided
	if (button.ItemMode==DDT_BUTTON) then local parent=button:GetParent(); while(parent.ddLevel and parent.ddLevel>=1) do parent:Hide(); parent=parent:GetParent(); end
	else button:GetParent():Redraw(); end
	button:Clicked(button.ItemChecked,(button.Extra.Edit and button.EditBox:GetText()) or nil);	-- func(button,check,editvalue)
end

-- also resets hide-counter for all parents
function DMMenu.ItemHover(button,enter)
	local parent=button:GetParent();
	if (not parent) then return; end
	parent.InOut[button]=enter;
	while(parent.ddLevel and parent.ddLevel>=1) do parent.elapsed=parent.ddTimer; parent=parent:GetParent(); end

	if (enter) then
		if (button.Tooltip) then
			if (type(button.Tooltip)=="function") then button:Tooltip(); return; end
			if (type(button.Tooltip)~="table") then button.Tooltip={ button.Tooltip }; end
			GameTooltip:SetOwner(button,"ANCHOR_LEFT");
			for i,text in ipairs(button.Tooltip) do
				if (i==1) then GameTooltip:SetText(button.Tooltip[i]);
				else GameTooltip:AddLine(button.Tooltip[i]); end
			end
			GameTooltip:Show();
		end
		if (button.ItemMode==DDT_MENU) then
			-- create and open sub-menu if present
			local frame=DMMenu:Create(button.Menu,button.Level+1,button);		-- create next level
			for _,item in ipairs(button.Menu) do frame:AddItem(item); end		-- add items (and show menu)
		end
	elseif (button.Tooltip) then GameTooltip:Hide(); end
end

function DMMenu.OnUpdate(frame,elapsed)
	frame:ReAlign();
	if (DMMenu.Frames[frame.ddLevel+1] and DMMenu.Frames[frame.ddLevel+1]:IsVisible()) then frame.elapsed=frame.ddTimer; return; end	-- visible child for whatever reason
	if (next(frame.InOut)) then frame.elapsed=frame.ddTimer; return; end		-- i have hover
	if (frame.NoTimeout) then frame.elapsed=frame.ddTimer; return; end		-- timeout halted for some other reason
	frame.elapsed=frame.elapsed-elapsed; if (frame.elapsed>0) then return; end	-- go timer
	frame.elapsed=-1;		-- clamp it
	frame:Hide();
	frame.ddAligned=nil;
	-- update any connected frames
	local i=frame.ddLevel+1;
	while(DMMenu.Frames[i]) do
		DMMenu.Frames[i]:Hide();
		DMMenu.Frames[i].ddAligned=nil;
		i=i+1;
	end
end

function DMMenu.frame_ReAlign(fr)
	if (fr.ddAligned or fr.ddEntries<1) then return; end
	local moved=false;
	local l,r,t,b=fr:GetLeft(),fr:GetRight(),fr:GetTop(),fr:GetBottom();
	-- submenu realignment
	if (fr.ddLevel>1) then
		if (r>fr.ddBounds.R) then												-- right edge OOB
			local skip=DMMenu.Frames[fr.ddLevel-1]:GetWidth()+fr:GetWidth()-2;	-- 2 pixels offset to lay edge on edge on the other side
			l=l-skip;
			r=r-skip;
			fr.ddPoint[4]=fr.ddPoint[4]-skip;
			moved=true;
		end
	end
	-- horisontal
	if (l<fr.ddBounds.L) then										-- left edge OOB
		fr.ddPoint[4]=fr.ddPoint[4]+(fr.ddBounds.L-l);				-- move without regard to right edge
		moved=true;
	elseif (r>fr.ddBounds.R) then									-- right edge OOB
		local maxmove,edgefault=l-fr.ddBounds.L,r-fr.ddBounds.R;	-- find limits
		if (edgefault>maxmove) then edgefault=maxmove; end			-- keep left edge visible
		fr.ddPoint[4]=fr.ddPoint[4]-edgefault;						-- move as appropriate
		if (edgefault~=0) then moved=true; end
	end
	-- vertical
	if (t>fr.ddBounds.T) then										-- top edge OOB
		fr.ddPoint[5]=fr.ddPoint[5]-(t-fr.ddBounds.T);				-- move without regard to bottom edge
		moved=true;
	elseif (b<fr.ddBounds.B) then									-- bottom edge OOB
		local maxmove,edgefault=fr.ddBounds.T-t,fr.ddBounds.B-b;	-- find limits
		if (edgefault>maxmove) then edgefault=maxmove; end			-- keep top edge visible
		fr.ddPoint[5]=fr.ddPoint[5]+edgefault;						-- move as appropriate
		if (edgefault~=0) then moved=true; end
	end
	-- effectuate
	if (moved) then
		fr:ClearAllPoints();
		fr:SetPoint(fr.ddPoint[1],fr.ddPoint[2],fr.ddPoint[3],fr.ddPoint[4],fr.ddPoint[5]);	-- spew point settings
	end
	fr.ddAligned=true;
end

-- drop-in replacement for native "CreateFrame"
function DMMenu.CreateFrame(base,name,parent,template)
	if (base=="CheckButton" and template=="DMMENUItem") then
		local widget=CreateFrame(base,name,parent);			-- basic base type
--		widget:SetBackdrop( { bgFile = "TILESET\\GENERIC\\Black", });
		widget:SetHeight(17);			-- <Size><AbsDimension y="17"/></Size>
		local icontexture=widget:CreateTexture(name.."IconTexture");
		icontexture:SetWidth(17); icontexture:SetHeight(17);
		icontexture:SetPoint("RIGHT");
--		icontexture:SetDrawLayer("BACKGROUND");
		local text=widget:CreateFontString(name.."Text");
		text:SetFontObject("GameFontHighlight");
		text:SetJustifyH("LEFT");
		text:SetPoint("LEFT",widget,"LEFT",3,0);
		text:SetPoint("RIGHT",widget,"RIGHT",0,0);
--		text:SetDrawLayer("ARTWORK");
		widget:SetNormalTexture("");
		widget:SetPushedTexture("Interface\\Buttons\\UI-Listbox-Highlight2");
		widget:SetCheckedTexture("Interface\\Buttons\\UI-Listbox-Highlight2","ADD");
		widget:SetHighlightTexture("Interface\\Buttons\\UI-Listbox-Highlight","ADD");
		widget:SetDisabledTexture("");
		-- OnLoad
		widget.IndentSize=19;
		widget.SetText=DMMenu.Item.SetButtonText;
		widget.GetText=DMMenu.Item.GetButtonText;
		widget.SetTextColor=DMMenu.Item.SetButtonTextColor;
		widget.SetIcon=DMMenu.Item.SetIcon;
		widget:RegisterForClicks("LeftButtonUp","RightButtonUp");
		-- other scripts
		widget:SetScript("PostClick",function () DMMenu.ItemClicked(widget); end);
		widget:SetScript("OnEnter",function () DMMenu.ItemHover(widget,true); end);
		widget:SetScript("OnLeave",function () DMMenu.ItemHover(widget,nil); end);
		return widget;
	end
	if (base=="Frame" and template=="DMMENU") then
		local cb1,cb2;
		local frame=CreateFrame(base,name,parent);			-- basic base type
		frame:SetFrameStrata("DIALOG");
		frame:SetBackdrop({ bgFile = "TILESET\\GENERIC\\Black", edgeFile="TILESET\\GENERIC\\Charcoal", edgeSize=1, });
		cb1=DMMenu.CreateFrame("CheckButton",name.."_Entry1",frame,"DMMENUItem");
		cb1:SetText("Entry 1"); cb1:SetPoint("TOPLEFT",frame,"TOPLEFT",1,-1); cb1:SetPoint("RIGHT",frame,"RIGHT",-1,0);
		cb2=DMMenu.CreateFrame("CheckButton",name.."_Entry2",frame,"DMMENUItem");
		cb2:SetText("Entry 2"); cb2:SetPoint("TOPLEFT",cb1,"BOTTOMLEFT",0,-1); cb2:SetPoint("RIGHT",cb1,"RIGHT");
		-- OnLoad
		frame.elapsed=3;				-- just so it doesn't crap out on some nil-math later
		frame.ddSpacing=1;
		frame.ddEntries=0;
		DMMenu.SetFuncs(frame);
		return frame;
	end

	return CreateFrame(base,name,parent,template);		-- standard type
end


end		-- DuckMod_Menu_Version < this version
